"""API tests"""
from email.utils import formataddr

import pytest
from anymail.message import AnymailMessage
from django.contrib.auth import get_user_model
from django.core.exceptions import ImproperlyConfigured

from mitol.mail.api import (
    build_message,
    can_email_user,
    get_connection,
    get_message_classes,
    get_message_sender,
    render_email_templates,
    safe_format_recipient,
    send_message,
)
from mitol.mail.test_utils import any_instance_of


pytestmark = [pytest.mark.django_db, pytest.mark.usefixtures("email_settings")]

User = get_user_model()


@pytest.fixture
def email_settings(settings):
    """Default settings for email tests"""
    settings.MITOL_MAIL_RECIPIENT_OVERRIDE = None


def test_get_message_classes(mocker, settings):
    """Verify get_message_classes dynamically imports named classes"""
    imported = [mocker.Mock(), mocker.Mock()]
    mock_import_string = mocker.patch(
        "mitol.mail.api.import_string", side_effect=imported
    )
    settings.MITOL_MAIL_MESSAGE_CLASSES = ["abc.Class1", "abc.Class2"]

    assert get_message_classes() == imported

    mock_import_string.assert_any_call("abc.Class1")
    mock_import_string.assert_any_call("abc.Class2")


def test_get_message_classes_invalid(mocker, settings):
    """Verify get_message_classes maps an import error into a ImproperlyConfigured error"""
    mocker.patch(
        "mitol.mail.api.import_string", side_effect=[mocker.Mock(), ImportError],
    )

    settings.MITOL_MAIL_MESSAGE_CLASSES = ["abc.Class1", "abc.Class2"]

    with pytest.raises(ImproperlyConfigured):
        get_message_classes()


def test_safe_format_email_recipient(mocker):
    """Test that safe_format_recipient uses the input if it's not a user"""
    assert safe_format_recipient("test@localhost") == "test@localhost"


@pytest.mark.parametrize("use_default", [True, False])
def test_safe_format_user_recipient(mocker, settings, use_default):
    """Test that safe_format_recipient calls the configured function if the recipient is a User instance"""
    if not use_default:
        settings.MITOL_MAIL_FORMAT_RECIPIENT_FUNC = "my.custom.function"

    mock_import_string = mocker.patch("mitol.mail.api.import_string")
    configured_func = mock_import_string.return_value
    user = mocker.Mock(
        spec=User, email="test@localhost", first_name="first", last_name="last"
    )
    assert safe_format_recipient(user) == mock_import_string.return_value.return_value

    configured_func.assert_called_once_with(user)
    mock_import_string.assert_called_once_with(
        "mitol.mail.defaults.format_recipient" if use_default else "my.custom.function"
    )


@pytest.mark.parametrize("use_default", [True, False])
def test_can_email_user(settings, mocker, use_default):
    """Test that can_email_user proxies to the configured function"""
    if not use_default:
        settings.MITOL_MAIL_CAN_EMAIL_USER_FUNC = "my.custom.function"

    mock_import_string = mocker.patch("mitol.mail.api.import_string")
    configured_func = mock_import_string.return_value
    user = mocker.Mock()
    assert can_email_user(user) == configured_func.return_value
    configured_func.assert_called_once_with(user)
    mock_import_string.assert_called_once_with(
        "mitol.mail.defaults.can_email_user" if use_default else "my.custom.function"
    )


def test_safe_format_recipient_override(mocker, settings):
    """Test that the recipient override works"""
    settings.MITOL_MAIL_RECIPIENT_OVERRIDE = "admin@localhost"
    assert safe_format_recipient(mocker.Mock()) == "admin@localhost"


def test_render_email_templates(mocker):
    """Test render_email_templates"""
    tmpls = {
        "test/subject.txt": "Welcome Jane Smith",
        "test/body.html": """<html><head>
<style type="text/css" data-premailer="ignore">
p {
  color: blue;
}
</style>
<style type="text/css">
.red {
  color: red;
}
</style>
</head><body>
<a href="http://example.com" class="red">html link</a>
</body></html>""",
    }
    mocker.patch(
        "mitol.mail.api.render_to_string", side_effect=lambda path, _: tmpls[path]
    )

    context = {"empty": False}

    subject, text_body, html_body = render_email_templates("test", context)
    assert subject == "Welcome Jane Smith"
    assert text_body == "html link (http://example.com)"
    assert html_body == (
        "<html><head>\n"
        '<style type="text/css">\n'
        "p {\n"
        "  color: blue;\n"
        "}\n"
        "</style>\n"
        "</head><body>\n"
        '<a href="http://example.com" class="red" style="color:red">html link</a>\n'
        "</body></html>"
    )


@pytest.mark.parametrize("recipient_is_user", [True, False])
def test_build_message(mocker, recipient_is_user):
    """
    Tests that build_message correctly builds a message object using the Anymail APIs
    """
    template_context = {"context_key": "context_value"}
    email = "user@localhost"
    recipient = (
        mocker.Mock(spec=User, email=email, first_name="First", last_name="Last")
        if recipient_is_user
        else email
    )
    mock_message_cls = mocker.Mock()
    mock_connection = mocker.Mock()

    assert (
        build_message(
            mock_connection,
            mock_message_cls,
            recipient,
            template_context,
            kwarg1=1,
            kwarg2=2,
        )
        == mock_message_cls.create.return_value
    )

    mock_message_cls.create.assert_called_once_with(
        connection=mock_connection,
        to=[safe_format_recipient(recipient)],
        template_context={
            **template_context,
            "user": recipient if recipient_is_user else None,
        },
        kwarg1=1,
        kwarg2=2,
    )


def test_send_message_none(mocker, mailoutbox):
    """Tests that send_message bails if there is no message"""
    patched_logger = mocker.patch("mitol.mail.api.log")

    send_message(None)

    patched_logger.exception.assert_not_called()


def test_send_message_failure(mocker, mailoutbox):
    """Tests that send_message logs all exceptions"""
    message = mocker.Mock()
    message.send.side_effect = ConnectionError
    patched_logger = mocker.patch("mitol.mail.api.log")

    send_message(message)

    message.send.assert_called_once()
    patched_logger.exception.assert_called_once()


@pytest.mark.parametrize("use_default", [True, False])
def test_get_connection(settings, mocker, use_default):
    if not use_default:
        settings.MITOL_MAIL_CONNECTION_BACKEND = "my.custom.backend"

    mock_django_get_connection = mocker.patch("mitol.mail.api.django_get_connection")

    with get_connection() as conn:
        assert conn == mock_django_get_connection.return_value

    mock_django_get_connection.assert_called_once_with(
        "anymail.backends.mailgun.EmailBackend" if use_default else "my.custom.backend"
    )


@pytest.mark.parametrize(
    "shared_context, expected_context",
    [({"a": 1}, {"a": 1, "b": 2}), ({}, {"b": 2}), (None, {"b": 2})],
)
def test_get_message_sender(mocker, shared_context, expected_context):
    """Test that get_message_sender configures a wrapper around sending APIs"""
    message_cls = mocker.Mock()
    recipient = mocker.Mock()
    mock_get_connection = mocker.patch("mitol.mail.api.get_connection")
    mock_build_message = mocker.patch("mitol.mail.api.build_message")
    mock_send_message = mocker.patch("mitol.mail.api.send_message")

    connection = mock_get_connection.return_value.__enter__.return_value

    with get_message_sender(message_cls, shared_context=shared_context) as sender:
        sender.build_and_send_message(recipient, {"b": 2}, tags=["tag1", "tag2"])

    mock_build_message.assert_called_once_with(
        connection, message_cls, recipient, expected_context, tags=["tag1", "tag2"]
    )
    mock_send_message.assert_called_once_with(mock_build_message.return_value)

    mock_get_connection.return_value.__enter__.assert_called_once_with()



"""Testing mail defaults"""

from mitol.mail.defaults import can_email_user, format_recipient


def test_format_recipient(mocker):
    """Test that format_recipient formats a user's email and name"""
    user = mocker.Mock(email="user@localhost", first_name="Sally", last_name="Ride")
    assert format_recipient(user) == "Sally Ride <user@localhost>"


def test_can_email_user(mocker):
    """Test that can_email_user returns True only if the user has an email"""
    assert can_email_user(mocker.Mock(email="user@localhost")) == True
    assert can_email_user(mocker.Mock(email="")) == False
    assert can_email_user(mocker.Mock(email=None)) == False



"""Tests for test utils"""
import pytest

from mitol.mail.test_utils import any_instance_of, assert_not_raises


def test_any_instance_of():
    """Tests any_instance_of()"""
    any_number = any_instance_of(int, float)

    assert any_number == 0.405
    assert any_number == 8_675_309
    assert any_number != "not a number"
    assert any_number != {}
    assert any_number != []


def test_assert_not_raises_none():
    """
    assert_not_raises should do nothing if no exception is raised
    """
    with assert_not_raises():
        pass


def test_assert_not_raises_exception(mocker):
    """assert_not_raises should fail the test"""
    # Here there be dragons
    fail_mock = mocker.patch("pytest.fail", autospec=True)
    with assert_not_raises():
        raise TabError()
    assert fail_mock.called is True


def test_assert_not_raises_failure():
    """assert_not_raises should reraise an AssertionError"""
    with pytest.raises(AssertionError):
        with assert_not_raises():
            assert 1 == 2



"""Testing utils"""
import abc
import traceback
from contextlib import contextmanager

import pytest


def any_instance_of(*cls):
    """
    Returns a type that evaluates __eq__ in isinstance terms

    Args:
        cls (list of types): variable list of types to ensure equality against

    Returns:
        AnyInstanceOf: dynamic class type with the desired equality
    """

    class AnyInstanceOf(metaclass=abc.ABCMeta):
        """Dynamic class type for __eq__ in terms of isinstance"""

        def __eq__(self, other):
            return isinstance(other, cls)

    for c in cls:
        AnyInstanceOf.register(c)
    return AnyInstanceOf()


@contextmanager
def assert_not_raises():
    """Used to assert that the context does not raise an exception"""
    try:
        yield
    except AssertionError:
        raise
    except Exception:  # pylint: disable=broad-except
        pytest.fail(f"An exception was not raised: {traceback.format_exc()}")



"""Message tests"""
import pytest

from mitol.mail.messages import TemplatedMessage


class NoNameMessage(TemplatedMessage):
    template_name = "testtemplatename"


class NoTemplateNameMessage(TemplatedMessage):
    name = "testname"


@pytest.fixture(autouse=True)
def default_settings(settings):
    """Default settings for tests"""
    settings.SITE_BASE_URL = "http://mit.edu/"
    settings.SITE_NAME = "MIT"
    settings.MITOL_MAIL_REPLY_TO_ADDRESS = "user@localhost"


def test_static_methods(settings):
    """Test TemplatedMessage's static methods"""
    assert TemplatedMessage.get_base_template_context() == {
        "base_url": settings.SITE_BASE_URL,
        "site_name": settings.SITE_NAME,
    }
    assert TemplatedMessage.get_debug_template_context() == {}
    assert TemplatedMessage.get_default_headers() == {
        "Reply-To": settings.MITOL_MAIL_REPLY_TO_ADDRESS
    }


def test_create_validations():
    """Verify the create() function validates the subclass setup"""
    with pytest.raises(ValueError) as exc_info:
        NoNameMessage.create()

    assert exc_info.value.args[0] == "NoNameMessage.name not defined"

    with pytest.raises(ValueError) as exc_info:
        NoTemplateNameMessage.create()

    assert exc_info.value.args[0] == "NoTemplateNameMessage.template_name not defined"



"""Mail forms"""
from django import forms

from mitol.mail.api import get_message_classes


def _get_message_class_choices():
    return [
        (message_cls.template_name, message_cls.name)
        for message_cls in get_message_classes()
    ]


class EmailDebuggerForm(forms.Form):
    """Form for email debugger"""

    template_name = forms.ChoiceField(
        choices=_get_message_class_choices,
        widget=forms.Select(attrs={"class": "form-control m-2"}),
    )
